unit COMportU;

// Please, don't delete this comment. \\
(*
  Copyright Owner: coltishWARE (Kevin Niehage)
  Copyright Year : 2008

  Unit   : COMportU (platform dependant)
  Version: 1.2c

  Contact E-Mail            : development@coltishWARE.com
  Contact Mobile            : +49-(0)-177-69-4321-8
  Contact Telephone Berlin  : +49-(0)-30-69-20-96-78
  Contact Telephone Cottbus : +49-(0)-355-28925-88-17-55
  Contact Telephone Mannheim: +49-(0)-621-483-4664
*)
// Please, don't delete this comment. \\

(*
  Description:

  This unit contains an interface to COM ports (/serial ports).
*)

(*
  Change Log:

  [Version 1.2c] (26.08.2008: sender release)
  - all events now provide a Sender parameter

  [Version 1.1c] (19.07.2008: synchronize release)
  - TCOMport is now synchronizable
  - buffer can now be flushed after being opened

  [Version 1.0c] (16.07.2008: initial release)
  - initial source has been written
*)

interface

uses
  Classes,
  SyncObjs,
  SysUtils,
  Windows;

const
  COMportU_CopyrightOwner = 'coltishWARE (Kevin Niehage)';
  COMportU_CopyrightYear  = '2008';
  COMportU_Name           = 'COMport';
  COMportU_ReleaseDate    = '26.08.2008';
  COMportU_ReleaseName    = 'sender release';
  COMportU_Version        = '1.2c';

const
  CopyrightOwner = COMportU_CopyrightOwner;
  CopyrightYear  = COMportU_CopyrightYear;
  Name           = COMportU_Name;
  ReleaseDate    = COMportU_ReleaseDate;
  ReleaseName    = COMportU_ReleaseName;
  Version        = COMportU_Version;

type
  TCOMport     = class;
  TCOMportRead = class;

  TReadCharEvent        = procedure (const ASender : TObject; const AChar : Char) of Object;
  TReadLineEvent        = procedure (const ASender : TObject; const ALine : String) of Object;
  TSetupConnectionEvent = procedure (const ASender : TObject; const ACOMhandle : THandle; var AConfig : TCommConfig) of Object;

  TCOMport = class(TObject)
  private
  protected
    FClearBufferOnOpen  : Boolean;
    FConfig             : TCommConfig;
    FCriticalSection    : TCriticalSection;
    FFlushBufferOnClose : Boolean;
    FLineBuffer         : String;
    FLineEnd            : String;
    FNumber             : Byte;
    FRead               : TCOMportRead;
    FReadOnly           : Boolean;
    FShowSetupDialog    : Boolean;

    FReadLineEvent        : TReadLineEvent;
    FSetupConnectionEvent : TSetupConnectionEvent;

    function GetSynchronized : Boolean;

    procedure ReadChar(const ASender : TObject; const AChar : Char);
    procedure SetConfig(const AValue : TCommConfig);
    procedure SetNumber(const AValue : Byte);
    procedure SetReadOnly(const AValue : Boolean);
    procedure SetSynchronized(const AValue : Boolean);
  public
    constructor Create(const ANumber : Byte);

    destructor Destroy; override;

    property ClearBufferOnOpen  : Boolean     read FClearBufferOnOpen  write FClearBufferOnOpen;
    property Config             : TCommConfig read FConfig             write SetConfig;
    property FlushBufferOnClose : Boolean     read FFlushBufferOnClose write FFlushBufferOnClose;
    property LineEnd            : String      read FLineEnd            write FLineEnd;
    property Number             : Byte        read FNumber             write SetNumber;
    property ReadOnly           : Boolean     read FReadOnly           write SetReadOnly;
    property ShowSetupDialog    : Boolean     read FShowSetupDialog    write FShowSetupDialog;
    property Synchronized       : Boolean     read GetSynchronized     write SetSynchronized;

    property OnReadLine        : TReadLineEvent        read FReadLineEvent        write FReadLineEvent;
    property OnSetupConnection : TSetupConnectionEvent read FSetupConnectionEvent write FSetupConnectionEvent;

    function CloseConnection : Boolean;
    function OpenConnection : Boolean;
  published
  end;

  TCOMportRead = class(TThread)
  private
  protected
    FCOMhandle : THandle;
    FCOMname   : String;

    FReadCharEvent : TReadCharEvent;

    procedure Execute; override;
  public
    constructor Create(const ACreateSuspended : Boolean);

    destructor Destroy; override;

    property COMhandle : THandle read FCOMhandle;
    property COMname   : String  read FCOMname;

    property OnReadChar : TReadCharEvent read FReadCharEvent write FReadCharEvent;

    function CloseCOMhandle : Boolean;
    function OpenCOMhandle(const ACOMname : String; const AReadOnly : Boolean; const APurgeBuffers : Boolean; var AConfig : TCommConfig) : Boolean;
  published
  end;

const
  CCOMport = '\\.\COM';

implementation

{ TCOMport }

function TCOMport.CloseConnection : Boolean;
begin
  Result := false;

  if (FRead <> nil) then
  begin
    if (Length(FLineBuffer) > 0) then
    begin
      if FFlushBufferOnClose then
      begin
        if Assigned(FReadLineEvent) then
          FReadLineEvent(Self, FLineBuffer);
      end;
    end;
    FLineBuffer := '';

    Result := FRead.CloseCOMhandle;
    if Result then
    begin
      FRead.Terminate;
      FRead.WaitFor;
      FRead.Free;
      FRead := nil;
    end;
  end;
end;

constructor TCOMport.Create(const ANumber: Byte);
begin
  inherited Create;

  FClearBufferOnOpen    := true;
  FCriticalSection      := nil;
  FLineBuffer           := '';
  FLineEnd              := #13#10;
  FNumber               := ANumber;
  FReadOnly             := false;
  FSetupConnectionEvent := nil;
  FShowSetupDialog      := true;

  OnReadLine        := nil;
  OnSetupConnection := nil;

  FConfig.dcb.DCBlength := SizeOf(FConfig.dcb);
  FConfig.dwSize        := SizeOf(FConfig);
end;

destructor TCOMport.Destroy;
begin
  CloseConnection;

  if (FCriticalSection <> nil) then
  begin
    FCriticalSection.Free;
    FCriticalSection := nil;
  end;

  inherited Destroy;
end;

function TCOMport.GetSynchronized : Boolean;
begin
  Result := (FCriticalSection <> nil);
end;

function TCOMport.OpenConnection : Boolean;
begin
  Result := false;

  if (FRead = nil) then
  begin
    FRead := TCOMportRead.Create(true);

    Result := FRead.OpenCOMhandle(CCOMport + IntToStr(FNumber), FReadOnly, FClearBufferOnOpen, FConfig);
    if Result then
    begin
      FLineBuffer := '';

      if FShowSetupDialog then
        CommConfigDialog(PChar(FRead.COMname), 0, FConfig);

      if Assigned(FSetupConnectionEvent) then
        FSetupConnectionEvent(Self, FRead.COMhandle, FConfig);

      SetConfig(FConfig);

      FRead.OnReadChar := ReadChar;
      FRead.Resume;
    end
    else
    begin
      FRead.Free;
      FRead := nil;
    end;
  end;
end;

procedure TCOMport.ReadChar(const ASender : TObject; const AChar : Char);
var
  LLine : String;
  LSync : TCriticalSection;
begin
  LSync := FCriticalSection;

  if (LSync <> nil) then
    LSync.Enter;
  try
    if Assigned(FReadLineEvent) then
    begin
      if (Length(FLineEnd) > 0) then
      begin
        FLineBuffer := FLineBuffer + AChar;
        while (Pos(FLineEnd, FLineBuffer) > 0) do
        begin
          LLine := Copy(FLineBuffer, 1, Pred(Pos(FLineEnd, FLineBuffer)));
          Delete(FLineBuffer, 1, Pos(FLineEnd, FLineBuffer) + Pred(Length(FLineEnd)));

          FReadLineEvent(Self, LLine);
        end;
      end
      else
        FReadLineEvent(Self, AChar);
    end;
  finally
    if (LSync = FCriticalSection) and (LSync <> nil) then
      LSync.Leave;
  end;
end;

procedure TCOMport.SetConfig(const AValue : TCommConfig);
begin
  if (FRead <> nil) then
    SetCommConfig(FRead.COMhandle, FConfig, SizeOf(FConfig));
end;

procedure TCOMport.SetNumber(const AValue: Byte);
begin
  if (FRead = nil) then
    FNumber := AValue;
end;

procedure TCOMport.SetReadOnly(const AValue : Boolean);
begin
  if (FRead = nil) then
    FReadOnly := AValue;
end;

procedure TCOMport.SetSynchronized(const AValue: Boolean);
begin
  if (FCriticalSection <> nil) then
  begin
    if not(AValue) then
    begin
      FCriticalSection.Free;
      FCriticalSection := nil;
    end;
  end
  else
  begin
    if AValue then
      FCriticalSection := TCriticalSection.Create;
  end;
end;

{ TCOMportRead }

function TCOMportRead.CloseCOMhandle: Boolean;
begin
  Result := false;

  if (FCOMhandle <> INVALID_HANDLE_VALUE) then
  begin
    Result := CloseHandle(FCOMhandle);

    FCOMhandle := INVALID_HANDLE_VALUE;
    FCOMname   := '';
  end;
end;

constructor TCOMportRead.Create(const ACreateSuspended : Boolean);
begin
  inherited Create(ACreateSuspended);

  FCOMhandle := INVALID_HANDLE_VALUE;
  FCOMname   := '';
end;

destructor TCOMportRead.Destroy;
begin
  CloseCOMhandle;

  inherited;
end;

procedure TCOMportRead.Execute;
var
  LChar  : Char;
  LRead  : LongWord;
begin
  while not(Terminated) do
  begin
    if (FCOMhandle <> INVALID_HANDLE_VALUE) then
    begin
      ReadFile(FCOMhandle, LChar, 1, LRead, nil);
      if (LRead = 1) then
      begin
        if Assigned(FReadCharEvent) then
          FReadCharEvent(Self, LChar);
      end;
    end;
  end;
end;

function TCOMportRead.OpenCOMhandle(const ACOMname: String; const AReadOnly : Boolean; const APurgeBuffers : Boolean; var AConfig : TCommConfig): Boolean;
var
  LAccess   : LongWord;
  LSize     : LongWord;
  LTimeouts : TCommTimeouts;
begin
  Result := false;

  if (FCOMhandle = INVALID_HANDLE_VALUE) then
  begin
    LAccess := GENERIC_READ;
    if not(AReadOnly) then
      LAccess := LAccess or GENERIC_WRITE;

    FCOMhandle := CreateFile(PChar(ACOMname),
                             LAccess,
                             0, // port not shareable
                             nil,
                             OPEN_EXISTING, // necessary for serial port
                             FILE_ATTRIBUTE_NORMAL,
                             0);
    FCOMname := ACOMname;

    Result := (FCOMhandle <> INVALID_HANDLE_VALUE);
    if Result then
    begin
      Result := GetCommTimeouts(FCOMhandle, LTimeouts);
      if Result then
      begin
        LTimeouts.ReadIntervalTimeout         := 0;
        LTimeouts.ReadTotalTimeoutMultiplier  := 0;
        LTimeouts.ReadTotalTimeoutConstant    := 0;
        LTimeouts.WriteTotalTimeoutMultiplier := 0;
        LTimeouts.WriteTotalTimeoutConstant   := 0;

        Result := SetCommTimeouts(FCOMhandle, LTimeouts);
        if Result then
        begin
          if APurgeBuffers then
            Result := PurgeComm(FCOMhandle, PURGE_RXCLEAR or PURGE_TXCLEAR or PURGE_RXABORT or PURGE_TXABORT);

          if Result then
            Result := GetCommConfig(FCOMhandle, AConfig, LSize);
        end;
      end;

      if not(Result) then
        CloseCOMhandle;
    end;
  end;
end;

end.
