unit pnpshr;

(*
March 13th, 2007
Kevin Niehage
(coltishWARE)

Definition of PortName Protocol (PNP)
-------------------------------------

PNP allows a client to request valid port
numbers via self-defined port names.


Default Port:
15937


Commands:
DISCONNECT
DCON
DC

HELP
H

RETURN <request>
RET    <request>
R      <request>

RETURNLIST [<request>]
RETL       [<request>]
RL         [<request>]


Replies:
100 DONE
200 OK
300 TEMPORARY ERROR
400 PERMANENT ERROR
401 UNKNOWN COMMAND
402 UNKNOWN PORTNAME
500 INTERNAL ERROR


Responses:
.<count> <response>


Communication Example:
(C = Client ; S = Server)

> Server listens on Port (e.g. 15937)
> Client connects to Server

C: RETURNLIST *FTP*
S: 200 OK
S: .1 TFTP1
S: .2 TFTP2
S: .3 FTP
S: .4 ANONYMOUS-FTP
S: 100 DONE
C: RETURN ANON-FTP
S: 200 OK
S: 402 UNKNOWN PORTNAME
C: RETURN ANONYMOUS-FTP
S: 200 OK
S: .1 21 ANONYMOUS-FTP
S: 100 DONE
C: RETURN *FTP*
S: 200 OK
S: .1 30 TFTP1
S: .2 31 TFTP2
S: .3 32 FTP
S: .4 21 ANONYMOUS-FTP
S: 100 DONE
c: DISCONNECT
S: 200 OK

> Client disconnects from Server
> Server closes connection

[EOF]
*)

interface

uses
  SysUtils;

// default port
const
  CPNPDefaultPort = 15937;

type
  TStringArray = array of String;

function AddToPortNameList(const APortName : String; const APortNumber : String) : Boolean;
function ClearPortNameList : Boolean;
function LoadPortNameList(const AFileName : String) : Boolean;
function SavePortNameList(const AFileName : String) : Boolean;

function HandleRequest(const ARequest : String; out ADisconnect : Boolean) : TStringArray;

implementation

const
  CDisConnectA = 'DISCONNECT';
  CDisConnectB = 'DCON';
  CDisConnectC = 'DC';

  CHelpA = 'HELP';
  CHelpB = 'H';

  CReturnA = 'RETURN';
  CReturnB = 'RET';
  CReturnC = 'R';

  CReturnListA = 'RETURNLIST';
  CReturnListB = 'RETL';
  CReturnListC = 'RL';

const
  CReply100 = '100 DONE';
  CReply200 = '200 OK';
  CReply300 = '300 TEMPORARY ERROR';
  CReply400 = '400 PERMANENT ERROR';
  CReply401 = '401 UNKNOWN COMMAND';
  CReply402 = '402 UNKNOWN PORTNAME';
  CReply500 = '500 INTERNAL ERROR';

const
  CResponsePrefix = '.';

const
  CParameterDivider  = ' ';
  CParameterEmbraceA = '''';
  CParameterEmbraceB = '"';

const
  CReplaceAny           = '*';
  CReplaceChoiceDivider = ',';
  CReplaceChoiceStart   = '[';
  CReplaceChoiceStop    = ']';
  CReplaceOne           = '?';

const
  CTextStart1 = 'A';
  CTextStart2 = '0';
  CTextStop1  = 'Z';
  CTextStop2  = '9';

type
  TBuildupEnum = (beNone, beChoice, beReplaceAny, beReplaceOne, beText);
  TCommandEnum = (ceNone, ceDisConnect, ceHelp, ceReturn, ceReturnList);
  TReplyEnum   = (reNone, reBusy, reFinished, reGeneralError, reUnderstood, reUnknownCommand, reUnknownPortName);

  TBuildupParseRecord = record
    Buildup    : TBuildupEnum;
    Parameters : TStringArray;
  end;
  TBuildupParseArray = array of TBuildupParseRecord;

  TCommandParseRecord = record
    Command    : TCommandEnum;
    Parameters : TStringArray;
  end;

  TPortNameRecord = record
    PortName   : String;
    PortNumber : String;
  end;
  TPortNameArray = array of TPortNameRecord;

  TLoopData = record
    Index : LongInt;
    Start : LongInt;
    Stop  : LongInt;
  end;
  TLoopDataArray = array of TLoopData;

  TLoopMethod = function (const ABuildup : TBuildupParseArray; const APortNames : TPortNameArray; var AResult : TPortNameArray; const ALoopDataArray : TLoopDataArray) : Boolean;

var
  VPortNameList : TPortNameArray;

function GetReply(const AReply : TReplyEnum) : String; forward;
function GetResponse(const AResponseNumber : LongWord; const AResponseText : String) : String; forward;

function FindPortNames(APortNames : TPortNameArray; ABuildup : TBuildupParseArray) : TPortNameArray; forward;
function MultiLoop(ALoopDataArray : TLoopDataArray; const ABuildup : TBuildupParseArray; const APortNames : TPortNameArray; var AResult : TPortNameArray; const ALoopMethod : TLoopMethod) : Boolean; forward;
function TryNextPortName(const ABuildup : TBuildupParseArray; const APortNames : TPortNameArray; var AResult : TPortNameArray; const ALoopDataArray : TLoopDataArray) : Boolean; forward;

function ParseBuildup(ABuildup : String) : TBuildupParseArray; forward;

function GetCommandEnumByText(const ACommand : String) : TCommandEnum; forward;
function ParseCommand(ACommand : String) : TCommandParseRecord; forward;

function GetReply(const AReply : TReplyEnum) : String;
begin
  case AReply of
    reBusy            : Result := CReply300;
    reFinished        : Result := CReply100;
    reGeneralError    : Result := CReply400;
    reUnderstood      : Result := CReply200;
    reUnknownCommand  : Result := CReply401;
    reUnknownPortName : Result := CReply402;
  else
    Result := CReply500;
  end;
end;

function GetResponse(const AResponseNumber : LongWord; const AResponseText : String) : String;
begin
  Result := CResponsePrefix + IntToStr(AResponseNumber) + CParameterDivider + AResponseText;
end;

function FindPortNames(APortNames : TPortNameArray; ABuildup : TBuildupParseArray) : TPortNameArray;
var
  LIndexA   : LongInt;
  LIndexB   : LongInt;
  LLoopData : TLoopDataArray;
begin
  SetLength(Result, 0);

  for LIndexA := 0 to Pred(Length(APortNames)) do
    APortNames[LIndexA].PortName := AnsiUpperCase(Trim(APortNames[LIndexA].PortName));

  SetLength(LLoopData, 0);
  for LIndexA := 0 to Pred(Length(ABuildup)) do
  begin
    if (ABuildup[LIndexA].Buildup = beChoice) then
    begin
      SetLength(LLoopData, Succ(Length(LLoopData)));

      LLoopData[High(LLoopData)].Start := Low(ABuildup[LIndexA].Parameters);
      LLoopData[High(LLoopData)].Stop  := High(ABuildup[LIndexA].Parameters);
    end;

    for LIndexB := 0 to Pred(Length(ABuildup[LIndexA].Parameters)) do
      ABuildup[LIndexA].Parameters[LIndexB] := AnsiUpperCase(Trim(ABuildup[LIndexA].Parameters[LIndexB]));
  end;

  if (Length(LLoopData) > 0) then
    MultiLoop(LLoopData, ABuildup, APortNames, Result, TryNextPortName)
  else
    TryNextPortName(ABuildup, APortNames, Result, LLoopData);
end;

function MultiLoop(ALoopDataArray : TLoopDataArray; const ABuildup : TBuildupParseArray; const APortNames : TPortNameArray; var AResult : TPortNameArray; const ALoopMethod : TLoopMethod) : Boolean;
var
  LIndex     : LongInt;
  LLoopLevel : LongInt;
begin
  Result := true;

  if ((Length(ALoopDataArray) > 0) and Assigned(ALoopMethod)) then
  begin
    for LIndex := Low(ALoopDataArray) to High(ALoopDataArray) do
      ALoopDataArray[LIndex].Index := ALoopDataArray[LIndex].Start;

    LLoopLevel := High(ALoopDataArray);
    while (LLoopLevel >= Low(ALoopDataArray)) do
    begin
      Result := Result and ALoopMethod(ABuildup, APortNames, AResult, ALoopDataArray);
      if not(Result) then
         Break;

      while (ALoopDataArray[LLoopLevel].Index >= ALoopDataArray[LLoopLevel].Stop) do
      begin
        Dec(LLoopLevel);

        if (LLoopLevel < Low(ALoopDataArray)) then
          Break;
      end;
      if (LLoopLevel >= Low(ALoopDataArray)) then
      begin
        Inc(ALoopDataArray[LLoopLevel].Index);
        for LIndex := Succ(LLoopLevel) to High(ALoopDataArray) do
          ALoopDataArray[LIndex].Index := ALoopDataArray[LIndex].Start;
        LLoopLevel := High(ALoopDataArray);
      end;
    end;
  end;
end;

function TryNextPortName(const ABuildup : TBuildupParseArray; const APortNames : TPortNameArray; var AResult : TPortNameArray; const ALoopDataArray : TLoopDataArray) : Boolean;
type
  TDifferenceEnum = (deNone, deAny, deOne);
var
  LCount           : LongInt;
  LCurrentPosition : LongInt;
  LCurrentText     : String;
  LDifference      : TDifferenceEnum;
  LError           : Boolean;
  LIndexA          : LongInt;
  LIndexB          : LongInt;
  LPortName        : String;
  LStringArray     : TStringArray;
begin
  if (Length(ALoopDataArray) > 0) then
  begin
    SetLength(LStringArray, Length(ALoopDataArray));
    for LIndexA := 0 to Pred(Length(ALoopDataArray)) do
    begin
      LCount := 0;
      for LIndexB := 0 to Pred(Length(ABuildup)) do
      begin
        if (ABuildup[LIndexB].Buildup = beChoice) then
        begin
          Inc(LCount);

          if (LCount = Succ(LIndexA)) then
          begin
            if (Length(ABuildup[LIndexB].Parameters) > ALoopDataArray[LIndexA].Index) then
              LStringArray[LIndexA] := ABuildup[LIndexB].Parameters[ALoopDataArray[LIndexA].Index];

            Break;
          end
        end;
      end;
    end;
  end;

  for LIndexA := 0 to Pred(Length(APortNames)) do
  begin
    LCount      := 0;
    LDifference := deNone;
    LError      := false;
    LPortName   := APortNames[LIndexA].PortName;
    for LIndexB := 0 to Pred(Length(ABuildup)) do
    begin
      case ABuildup[LIndexB].Buildup of
        beChoice, beText :
        begin
          if (ABuildup[LIndexB].Buildup = beChoice) then
          begin
            Inc(LCount);
            LError := (LCount > Length(LStringArray));
            if not(LError) then
              LCurrentText := LStringArray[Pred(LCount)];
          end
          else
          begin
            LError := (Length(ABuildup[LIndexB].Parameters) <= 0);
            if not(LError) then
              LCurrentText := ABuildup[LIndexB].Parameters[0];
          end;

          if not(LError) then
          begin
            LCurrentPosition := Pos(LCurrentText, LPortName);
            case LDifference of
              deNone : LError := (LCurrentPosition <> 1);
              deOne  : LError := ((LCurrentPosition <> 1) and (LCurrentPosition <> 2));
            else
              LError := ((LDifference <> deAny) or (LCurrentPosition <= 0));
            end;

            if not(LError) then
            begin
              LDifference := deNone;
              Delete(LPortName, 1, Pred(LCurrentPosition + Length(LCurrentText)));
            end;
          end;
        end;

        beReplaceAny :
        begin
          LDifference := deAny;
        end;

        beReplaceOne :
        begin
          LDifference := deOne;
        end;
      else
        LError := true;
      end;

      if LError then
        Break;
    end;

    if not(LError) then
    begin
      SetLength(AResult, Succ(Length(AResult)));
      AResult[High(AResult)] := APortNames[LIndexA];
    end;
  end;

  Result := true;
end;

function ParseBuildup(ABuildup : String) : TBuildupParseArray;
type
  TProgressEnum = (peNone, peChoice, peText);
var
  LCount    : LongInt;
  LCurrent  : String;
  LError    : Boolean;
  LIndex    : LongInt;
  LProgress : TProgressEnum;
begin
  SetLength(Result, 0);

  ABuildup := AnsiUpperCase(Trim(ABuildup));

  LCount    := 0;
  LError    := false;
  LProgress := peNone;
  SetLength(LCurrent, Length(ABuildup));
  for LIndex := 1 to Length(ABuildup) do
  begin
    case ABuildup[LIndex] of
      CReplaceAny :
      begin
        // important that LProgress is set to peNone
        if (LProgress = peText) then
        begin
          if (LCount > 0) then
          begin
            SetLength(LCurrent, LCount);

            SetLength(Result, Succ(Length(Result)));
            SetLength(Result[High(Result)].Parameters, 1);

            Result[High(Result)].Buildup := beText;
            Result[High(Result)].Parameters[High(Result[High(Result)].Parameters)] := LCurrent;

            LCount := 0;

            SetLength(LCurrent, Length(ABuildup) - LIndex);
          end;

          LProgress := peNone;
        end;

        if (LProgress = peNone) then
        begin
          SetLength(Result, Succ(Length(Result)));
          SetLength(Result[High(Result)].Parameters, 0);

          Result[High(Result)].Buildup := beReplaceAny;

          LCount := 0;
        end
        else
          LError := true;
      end;

      CReplaceChoiceDivider :
      begin
        if (LProgress = peChoice) then
        begin
          SetLength(LCurrent, LCount);

          SetLength(Result[High(Result)].Parameters, Succ(Length(Result[High(Result)].Parameters)));

          Result[High(Result)].Buildup := beChoice;
          Result[High(Result)].Parameters[High(Result[High(Result)].Parameters)] := LCurrent;

          LCount := 0;

          SetLength(LCurrent, Length(ABuildup) - LIndex);
        end
        else
          LError := true;
      end;

      CReplaceChoiceStart :
      begin
        // important that LProgress is set to peNone
        if (LProgress = peText) then
        begin
          if (LCount > 0) then
          begin
            SetLength(LCurrent, LCount);

            SetLength(Result, Succ(Length(Result)));
            SetLength(Result[High(Result)].Parameters, 1);

            Result[High(Result)].Buildup := beText;
            Result[High(Result)].Parameters[High(Result[High(Result)].Parameters)] := LCurrent;

            LCount := 0;

            SetLength(LCurrent, Length(ABuildup) - LIndex);
          end;

          LProgress := peNone;
        end;

        if (LProgress = peNone) then
        begin
          SetLength(Result, Succ(Length(Result)));
          SetLength(Result[High(Result)].Parameters, 0);

          Result[High(Result)].Buildup := beChoice;

          LCount    := 0;
          LProgress := peChoice;
        end;
      end;

      CReplaceChoiceStop :
      begin
        if (LProgress = peChoice) then
        begin
          SetLength(LCurrent, LCount);

          SetLength(Result[High(Result)].Parameters, Succ(Length(Result[High(Result)].Parameters)));

          Result[High(Result)].Buildup := beChoice;
          Result[High(Result)].Parameters[High(Result[High(Result)].Parameters)] := LCurrent;

          LCount    := 0;
          LProgress := peNone;

          SetLength(LCurrent, Length(ABuildup) - LIndex);
        end
        else
          LError := true;
      end;

      CReplaceOne :
      begin
        // important that LProgress is set to peNone
        if (LProgress = peText) then
        begin
          if (LCount > 0) then
          begin
            SetLength(LCurrent, LCount);

            SetLength(Result, Succ(Length(Result)));
            SetLength(Result[High(Result)].Parameters, 1);

            Result[High(Result)].Buildup := beText;
            Result[High(Result)].Parameters[High(Result[High(Result)].Parameters)] := LCurrent;

            LCount := 0;

            SetLength(LCurrent, Length(ABuildup) - LIndex);
          end;

          LProgress := peNone;
        end;

        if (LProgress = peNone) then
        begin
          SetLength(Result, Succ(Length(Result)));
          SetLength(Result[High(Result)].Parameters, 0);

          Result[High(Result)].Buildup := beReplaceOne;

          LCount := 0;
        end
        else
          LError := true;
      end;

      CTextStart1 .. CTextStop1, CTextStart2 .. CTextStop2 :
      begin
        if (LProgress = peNone) then
          LProgress := peText;

        Inc(LCount);
        LCurrent[LCount] := ABuildup[LIndex];
      end;
    else
      LError := true;
    end;

    if LError then
      Break;
  end;

  if ((LCount > 0) and not(LError)) then
  begin
    case LProgress of
      peText :
      begin
        SetLength(LCurrent, LCount);

        SetLength(Result, Succ(Length(Result)));
        SetLength(Result[High(Result)].Parameters, 1);

        Result[High(Result)].Buildup := beText;
        Result[High(Result)].Parameters[High(Result[High(Result)].Parameters)] := LCurrent;
      end;
    else
      LError := true;
    end;
  end;

  if LError then
  begin
    for LIndex := 0 to Pred(Length(Result)) do
      SetLength(Result[LIndex].Parameters, 0);
    SetLength(Result, 0);
  end;
end;

function GetCommandEnumByText(const ACommand : String) : TCommandEnum;
begin
  Result := ceNone;

  if ((ACommand = CDisConnectA) or (ACommand = CDisConnectB) or (ACommand = CDisConnectC)) then
    Result := ceDisConnect;
  if ((ACommand = CHelpA) or (ACommand = CHelpB)) then
    Result := ceHelp;
  if ((ACommand = CReturnA) or (ACommand = CReturnB) or (ACommand = CReturnC)) then
    Result := ceReturn;
  if ((ACommand = CReturnListA) or (ACommand = CReturnListB) or (ACommand = CReturnListC)) then
    Result := ceReturnList;
end;

function ParseCommand(ACommand : String) : TCommandParseRecord;
type
  TProgressEnum = (peNone, peCommand, peCommandEmbrace, peParameter, peParameterEmbrace);
const
  CEmptyEmbrace = #0;
var
  LCount    : LongInt;
  LCurrent  : String;
  LEmbrace  : Char;
  LError    : Boolean;
  LIndex    : LongInt;
  LProgress : TProgressEnum;
begin
  Result.Command := ceNone;
  SetLength(Result.Parameters, 0);

  ACommand := AnsiUpperCase(Trim(ACommand));

  LCount    := 0;
  LEmbrace  := CEmptyEmbrace;
  LError    := false;
  LProgress := peCommand;
  SetLength(LCurrent, Length(ACommand));
  for LIndex := 1 to Length(ACommand) do
  begin
    case ACommand[LIndex] of
      CParameterDivider :
      begin
        case LProgress of
          peCommand :
          begin
            if (LCount > 0) then
            begin
              SetLength(LCurrent, LCount);

              Result.Command := GetCommandEnumByText(LCurrent);
              LError := (Result.Command = ceNone);

              LCount    := 0;
              LEmbrace  := CEmptyEmbrace;
              LProgress := peParameter;

              SetLength(LCurrent, Length(ACommand) - LIndex);
            end
            else
              LError := true;
          end;

          peCommandEmbrace, peParameterEmbrace :
          begin
            Inc(LCount);
            LCurrent[LCount] := ACommand[LIndex];
          end;

          peParameter :
          begin
            if (LCount >= 0) then
            begin
              SetLength(LCurrent, LCount);

              SetLength(Result.Parameters, Succ(Length(Result.Parameters)));
              Result.Parameters[High(Result.Parameters)] := LCurrent;

              LCount    := 0;
              LEmbrace  := CEmptyEmbrace;
              LProgress := peParameter;

              SetLength(LCurrent, Length(ACommand) - LIndex);
            end;
          end;
        else
          LError := true;
        end;
      end;

      CParameterEmbraceA, CParameterEmbraceB :
      begin
        if ((LEmbrace <> CEmptyEmbrace) and (ACommand[LIndex] = LEmbrace)) then
        begin
          LEmbrace := CEmptyEmbrace;

          case LProgress of
            peCommandEmbrace :
            begin
              SetLength(LCurrent, LCount);

              Result.Command := GetCommandEnumByText(LCurrent);
              LError := (Result.Command = ceNone);

              LCount    := 0;
              LEmbrace  := CEmptyEmbrace;
              LProgress := peCommand;

              SetLength(LCurrent, Length(ACommand) - LIndex);
            end;

            peParameterEmbrace :
            begin
              SetLength(LCurrent, LCount);

              SetLength(Result.Parameters, Succ(Length(Result.Parameters)));
              Result.Parameters[High(Result.Parameters)] := LCurrent;

              LCount    := 0;
              LEmbrace  := CEmptyEmbrace;
              LProgress := peParameter;

              SetLength(LCurrent, Length(ACommand) - LIndex);
            end;
          else
            LError := true;
          end;
        end
        else
        begin
          if (LEmbrace = CEmptyEmbrace) then
          begin
            if (LCount <= 0) then
            begin
              LEmbrace := ACommand[LIndex];

              case LProgress of
                peCommand   : LProgress := peCommandEmbrace;
                peParameter : LProgress := peParameterEmbrace;
              else
                LError := true;
              end;
            end
            else
              LError := true;
          end
          else
          begin
            Inc(LCount);
            LCurrent[LCount] := ACommand[LIndex];
          end;
        end;
      end;

      CTextStart1 .. CTextStop1, CTextStart2 .. CTextStop2,
      CReplaceAny, CReplaceChoiceDivider, CReplaceChoiceStart, CReplaceChoiceStop, CReplaceOne :
      begin
        LError := ((ACommand[LIndex] in [CReplaceAny, CReplaceChoiceDivider, CReplaceChoiceStart, CReplaceChoiceStop, CReplaceOne]) and (LProgress = peCommand));
        if not(LError) then
        begin
          Inc(LCount);
          LCurrent[LCount] := ACommand[LIndex];
        end;
      end;
    else
      LError := true;
    end;

    if LError then
      Break;
  end;

  if ((LCount > 0) and not(LError)) then
  begin
    case LProgress of
      peCommand :
      begin
        SetLength(LCurrent, LCount);

        Result.Command := GetCommandEnumByText(LCurrent);
        LError := (Result.Command = ceNone);
      end;

      peParameter :
      begin
        SetLength(LCurrent, LCount);

        SetLength(Result.Parameters, Succ(Length(Result.Parameters)));
        Result.Parameters[High(Result.Parameters)] := LCurrent;
      end;
    else
      LError := true;
    end;
  end;

  if LError then
  begin
    Result.Command := ceNone;
    SetLength(Result.Parameters, 0);
  end;
end;

function AddToPortNameList(const APortName : String; const APortNumber : String) : Boolean;
var
  LOldLength : LongInt;
begin
  LOldLength := Length(VPortNameList);

  if ((Length(Trim(APortName)) > 0) and (Length(Trim(APortNumber)) > 0)) then
  begin
    SetLength(VPortNameList, Succ(Length(VPortNameList)));
    VPortNameList[High(VPortNameList)].PortName   := AnsiUpperCase(Trim(APortName));
    VPortNameList[High(VPortNameList)].PortNumber := AnsiUpperCase(Trim(APortNumber));
  end;

  Result := (Length(VPortNameList) = Succ(LOldLength));
end;

function ClearPortNameList : Boolean;
begin
  SetLength(VPortNameList, 0);

  Result := (Length(VPortNameList) = 0);
end;

function LoadPortNameList(const AFileName : String) : Boolean;
var
  LCount      : LongInt;
  LLine       : String;
  LLoadFile   : TextFile;
  LOldLength  : LongInt;
  LPortName   : String;
  LPortNumber : String;
begin
  Result := false;

  if FileExists(AFileName) then
  begin
    LCount     := 0;
    LOldLength := Length(VPortNameList);

    AssignFile(LLoadFile, AFileName);
    try
      Reset(LLoadFile);

      while not(EoF(LLoadFile)) do
      begin
        ReadLn(LLoadFile, LLine);

        if (Pos(CParameterDivider, LLine) > 0) then
        begin
          LPortName   := Copy(LLine, 1, Pred(Pos(CParameterDivider, LLine)));
          LPortNumber := Copy(LLine, Succ(Pos(CParameterDivider, LLine)), Length(LLine) - Pos(CParameterDivider, LLine));
        end
        else
        begin
          LPortName   := LLine;
          LPortNumber := '';
        end;

        if ((Length(Trim(LPortName)) > 0) and (Length(Trim(LPortNumber)) > 0)) then
        begin
          SetLength(VPortNameList, Succ(Length(VPortNameList)));
          VPortNameList[High(VPortNameList)].PortName   := AnsiUpperCase(Trim(LPortName));
          VPortNameList[High(VPortNameList)].PortNumber := AnsiUpperCase(Trim(LPortNumber));

          Inc(LCount);
        end;
      end;
    finally
      CloseFile(LLoadFile);
    end;

    Result := (Length(VPortNameList) = (LOldLength + LCount));
  end;
end;

function SavePortNameList(const AFileName : String) : Boolean;
var
  LCount    : LongInt;
  LIndex    : LongInt;
  LSaveFile : TextFile;
begin
  Result := false;

  LCount := 0;

  AssignFile(LSaveFile, AFileName);
  try
    Rewrite(LSaveFile);

    for LIndex := 0 to Pred(Length(VPortNameList)) do
    begin
      if ((Length(Trim(VPortNameList[LIndex].PortName)) > 0) and (Length(Trim(VPortNameList[LIndex].PortNumber)) > 0)) then
      begin
        WriteLn(LSaveFile, AnsiUpperCase(Trim(VPortNameList[LIndex].PortName)) + CParameterDivider + AnsiUpperCase(Trim(VPortNameList[LIndex].PortNumber)));

        Inc(LCount);
      end;
    end;
  finally
    CloseFile(LSaveFile);
  end;

  Result := (LCount = Length(VPortNameList));
end;

function HandleRequest(const ARequest : String; out ADisconnect : Boolean) : TStringArray;
var
  LBuildup   : TBuildupParseArray;
  LCommand   : TCommandParseRecord;
  LError     : Boolean;
  LIndex     : LongInt;
  LPortNames : TPortNameArray;
  LReplyA    : String;
  LReplyB    : String;
  LResponse  : TStringArray;
begin
  SetLength(Result, 0);
  ADisconnect := false;

  LReplyA := '';
  LReplyB := '';
  SetLength(LPortNames, 0);
  SetLength(LResponse, 0);

  try
    LCommand := ParseCommand(ARequest);
    case LCommand.Command of
      ceDisConnect :
      begin
        LReplyA := GetReply(reUnderstood);

        ADisconnect := true;
      end;

      ceHelp :
      begin
        LReplyA := GetReply(reUnderstood);

        SetLength(LResponse, 4);
        LResponse[0] := CDisConnectA + CParameterDivider + CDisConnectB + CParameterDivider + CDisConnectC;
        LResponse[1] := CHelpA + CParameterDivider + CHelpB;
        LResponse[2] := CReturnA + CParameterDivider + CReturnB + CParameterDivider + CReturnC;
        LResponse[3] := CReturnListA + CParameterDivider + CReturnListB + CParameterDivider + CReturnListC;

        LReplyB := GetReply(reFinished);
      end;

      ceReturn, ceReturnList :
      begin
        LReplyA := GetReply(reUnderstood);

        LError := false;
        SetLength(LBuildup, 0);
        if (Length(LCommand.Parameters) > 0) then
        begin
          LBuildup := ParseBuildup(LCommand.Parameters[0]);
          LError   := (Length(LBuildup) <= 0);
        end;

        if not(LError) then
        begin
          LPortNames := FindPortNames(VPortNameList, LBuildup);
          if (Length(LPortNames) > 0) then
          begin
            SetLength(LResponse, Length(LPortNames));
            for LIndex := 0 to Pred(Length(LPortNames)) do
            begin
              if (LCommand.Command = ceReturn) then
                LResponse[LIndex] := LPortNames[LIndex].PortNumber + CParameterDivider + LPortNames[LIndex].PortName
              else
                LResponse[LIndex] := LPortNames[LIndex].PortName;
            end;

            LReplyB := GetReply(reFinished);
          end
          else
            LReplyB := GetReply(reUnknownPortName);
        end
        else
          LReplyB := GetReply(reUnknownPortName);
      end;
    else
      LReplyA := GetReply(reUnknownCommand);
    end;
  except
    LReplyB := GetReply(reGeneralError);
  end;

  SetLength(Result, Succ(Length(LResponse)) + Ord(Length(LReplyB) > 0));
  if (Length(Result) > 0) then
  begin
    Result[0] := LReplyA;
    for LIndex := 0 to Pred(Length(LResponse)) do
    begin
      if (Length(Result) >= Succ(LIndex)) then
        Result[Succ(LIndex)] := GetResponse(Succ(LIndex), LResponse[LIndex]);
    end;

    if ((Length(LReplyB) > 0) and (Length(Result) >= (Succ(Length(LResponse)) + Ord(Length(LReplyB) > 0)))) then
      Result[Succ(Length(LResponse))] := LReplyB;
  end;
end;

end.
